import java.awt.*;
import java.util.*;
import java.awt.event.*;
import javax.swing.*;
import javax.media.opengl.*;
import javax.media.opengl.glu.*;

public class Geom implements Iterable<Geom.T> {

    private HashMap<P,P> ps = new HashMap<P,P>();
    private HashMap<E,E> es = new HashMap<E,E>();
    private HashSet<T>   ts = new HashSet<T>();

    public Iterator<T> iterator() { return ts.iterator(); }

    public P newP(float x, float y, float z) {
        P p = new P(x, y, z);
        P p2 = ps.get(p);
        if (p2 != null) return p2;
        ps.put(p,p);
        return p;
    }
    
    public E newE(P p1, P p2) {
        E e = new E(p1, p2);
        E e2 = es.get(e);
        if (e2 != null) return e2;
        es.put(e,e);
        return e;
    }

    public T newT(E e1, E e2, E e3) {
        if (e1.t1 != null && e1.t1.hasE(e1) && e1.t1.hasE(e2) && e1.t1.hasE(e3)) return e1.t1;
        if (e1.t2 != null && e1.t2.hasE(e1) && e1.t2.hasE(e2) && e1.t2.hasE(e3)) return e1.t2;
        if (e2.t1 != null && e2.t1.hasE(e1) && e2.t1.hasE(e2) && e2.t1.hasE(e3)) return e2.t1;
        if (e2.t2 != null && e2.t2.hasE(e1) && e2.t2.hasE(e2) && e2.t2.hasE(e3)) return e2.t2;
        if (e3.t1 != null && e3.t1.hasE(e1) && e3.t1.hasE(e2) && e3.t1.hasE(e3)) return e3.t1;
        if (e3.t2 != null && e3.t2.hasE(e1) && e3.t2.hasE(e2) && e3.t2.hasE(e3)) return e3.t2;
        return new T(e1, e2, e3);
    }

    /** [UNIQUE] point in 3-space */
    public final class P {
        public final float x, y, z;
        public P(float x, float y, float z) { this.x = x; this.y = y; this.z = z; }
        public V minus(P p) { return new V(x-p.x, y-p.y, z-p.z); }
        public boolean equals(Object o) {
            if (o==null || !(o instanceof P)) return false;
            P p = (P)o;
            return p.x==x && p.y==y && p.z==z;
        }
        public int hashCode() {
            return
                Float.floatToIntBits(x) ^
                Float.floatToIntBits(y) ^
                Float.floatToIntBits(z);
        }
        public void glVertex(GL gl) { gl.glVertex3f(x, y, z); }
    }

    /** vector in 3-space */
    public final class V {
        private final float x, y, z;
        public V(float x, float y, float z) { this.x = x; this.y = y; this.z = z; }
        public boolean sameDirection(V v) { throw new Error(); }
        public V cross(V v) { return new V(y*v.z-z*v.y, z*v.x-x*v.z, x*v.y-y*v.x); }
        public float mag() { return (float)Math.sqrt(x*x+y*y+z*z); }
        public V norm() { float m = mag(); return new V(x/m, y/m, z/m); }
    }

    /** [UNIQUE] an edge */
    public final class E {
        public final P p1, p2;
        T t1, t2;
        public E(P p1, P p2) { this.p1 = p1; this.p2 = p2; }
        public int hashCode() { return p1.hashCode() ^ p2.hashCode(); }
        public boolean equals(Object o) {
            if (o==null || !(o instanceof E)) return false;
            E e = (E)o;
            if (this.p1 == e.p1 && this.p2 == e.p2) return true;
            if (this.p2 == e.p1 && this.p1 == e.p2) return true;
            return false;
        }
        public P shared(E e) {
            if (p1==e.p1) return p1;
            if (p1==e.p2) return p1;
            if (p2==e.p1) return p2;
            if (p2==e.p2) return p2;
            throw new Error("no shared vertex in shared()");
        }
        public P unshared(E e) {
            if (p1==e.p1) return p2;
            if (p1==e.p2) return p2;
            if (p2==e.p1) return p1;
            if (p2==e.p2) return p1;
            throw new Error("no shared vertex in unshared()");
        }
    }

    /** [UNIQUE] a triangle (face) */
    public final class T {
        public final E e1, e2, e3;
        public T(E e1, E e2, E e3, V normal) {
            P p1 = e1.shared(e2);
            P p2 = e2.shared(e3);
            P p3 = e3.shared(e1);
            V norm = p2.minus(p1).cross(p3.minus(p1));
            if (norm.sameDirection(normal)) {
                this.e1 = e1;
                this.e2 = e2;
                this.e3 = e3;
            } else {
                this.e1 = e3;
                this.e2 = e2;
                this.e3 = e1;
            }
            if (e1.t1 == null) e1.t1 = this; else if (e1.t2 == null) e1.t2 = this; else throw new Error("non-manifold surface");
            if (e2.t1 == null) e2.t1 = this; else if (e2.t2 == null) e2.t2 = this; else throw new Error("non-manifold surface");
            if (e3.t1 == null) e3.t1 = this; else if (e3.t2 == null) e3.t2 = this; else throw new Error("non-manifold surface");
            ts.add(this);
        }
        public V norm() {
            P p1 = e1.shared(e2);
            P p2 = e2.shared(e3);
            P p3 = e3.shared(e1);
            return p2.minus(p1).cross(p3.minus(p1)).norm();
        }
        public T(E e1, E e2, E e3) {
            this.e1 = e1;
            this.e2 = e2;
            this.e3 = e3;
            if (e1.t1 == null) e1.t1 = this; else if (e1.t2 == null) e1.t2 = this; else throw new Error("non-manifold surface");
            if (e2.t1 == null) e2.t1 = this; else if (e2.t2 == null) e2.t2 = this; else throw new Error("non-manifold surface");
            if (e3.t1 == null) e3.t1 = this; else if (e3.t2 == null) e3.t2 = this; else throw new Error("non-manifold surface");
            ts.add(this);
        }
        public boolean hasE(E e) { return e1==e || e2==e || e3==e; }
        public void glVertices(GL gl) {
            e1.unshared(e2).glVertex(gl);
            e1.shared(e2).glVertex(gl);
            e2.shared(e3).glVertex(gl);
        }
    }

    /** matrix */
    public class M {
    }

}
